/*
 * Copyright (c) 2004 Apple Computer, Inc. All Rights Reserved.
 * 
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */


//
// child - track a single child process and its belongings
//
#ifndef _CHILD_H_
#define _CHILD_H_  1

#include <security_utilities/mach++.h>
#include <security_utilities/unixchild.h>

using MachPlusPlus::Port;
using MachPlusPlus::TaskPort;


//
// ServerChild builds on the generic UNIX Child abstraction.
// The child is expected to engage in a checkin protocol after launch,
// whereby it RPC-calls childCheckIn in securityd and thus authenticates
// and declares readiness to provide service.
//
// @@@ PerWhat are these, if they are at all?
//
class ServerChild : public UnixPlusPlus::Child {
public:
	ServerChild();
	~ServerChild();
	
	Port servicePort() const { return mServicePort; }
	bool ready() const { return mServicePort; }

public:
	static void checkIn(Port servicePort, pid_t pid);

protected:
	void childAction() = 0;		// must be provided by subclass
	void parentAction();		// fully implemented
	void dying();				// fully implemented

private:
	Port mServicePort;			// child's main service port

private:
	typedef map<pid_t, ServerChild *> CheckinMap;
	static CheckinMap mCheckinMap;

	// The parent side will wait on mCheckinCond until the child checks in
	// or fails. During that time ONLY, mCheckinLock protects the entire Child
	// object.
	static Mutex mCheckinLock;
	Condition mCheckinCond;
};


#endif // _CHILD_H_
